#!/usr/bin/env python

import os
import sys
import argparse
import logging

import filelock

try:
    import perfact.loggingtools
except ImportError:
    pass

from .helpers import load_config
from .zodbsync import ZODBSync

from .commands.record import Record
from .commands.playback import Playback
from .commands.watch import Watch
from .commands.pick import Pick
from .commands.upload import Upload
from .commands.with_lock import WithLock
from .commands.reset import Reset
from .commands.execute import Exec
from .commands.reformat import Reformat
from .commands.checkout import Checkout
from .commands.freeze import Freeze
from .commands.layer_hash import LayerHash
from .commands.layer_init import LayerInit
from .commands.layer_update import LayerUpdate


class Runner(object):
    """
    Parses arguments to select the correct SubCommand subclass.
    """
    commands = [Record, Playback, Watch, Pick, Upload, WithLock, Reset, Exec,
                Reformat, Checkout, Freeze, LayerHash, LayerInit, LayerUpdate]

    def __init__(self):
        """
        Set up the argument parser with the possible subcommands
        """
        parser = argparse.ArgumentParser(description='''
            Tool to sync objects between a ZODB and a git-controlled folder on
            the file system.
        ''')
        default_configfile = '/etc/perfact/modsync/zodb.py'
        parser.add_argument(
            '--config', '-c', type=str,
            help='Path to config (default: %s)' % default_configfile,
            default=default_configfile
        )
        parser.add_argument(
            '--no-lock', action='store_true',
            help='Do not acquire lock. Only use inside a with-lock wrapper.',
        )
        if 'perfact.loggingtools' in sys.modules:
            perfact.loggingtools.addArgs(parser, name='ZODBSync')

        # Add all available SubCommand classes as sub-command runners, using
        # either the property "subcommand" or the name of the class.
        # The chosen subcommand class will be available as args.command
        subs = parser.add_subparsers()
        for cls in self.commands:
            name = getattr(cls, 'subcommand', cls.__name__.lower())
            subparser = subs.add_parser(
                name,
                help=cls.__doc__,
            )
            cls.add_args(subparser)
            subparser.set_defaults(command=cls)

        self.parser = parser

        # These are set by parse()
        self.args = None
        self.logger = None
        self.config = None
        self.lock = None
        self.sync = None
        self.command = None

    def parse(self, *argv):
        """
        Parse the given arguments and set the command accordingly. If no
        arguments are given, sys.argv is used.
        This can be called several times on the same instance to re-use the
        connection, but the config needs to be the same.
        """
        args = self.parser.parse_args(argv if argv else None)
        self.args = args

        if 'perfact.loggingtools' in sys.modules:
            logger = perfact.loggingtools.createLogger(
                args=args, name='ZODBSync'
            )
        else:
            logger = logging.getLogger('ZODBSync')
            logger.setLevel(logging.INFO)
            logger.addHandler(logging.StreamHandler())
            logger.propagate = True

        self.logger = logger
        if getattr(args.command, 'use_config', True):
            config = load_config(args.config)
            if self.config is not None and config != self.config:
                self.logger.warning("Reusing runner with different config")
                self.sync = None
            self.config = config

        # Usually, each command needs a connection to the ZODB, but it might
        # explicitly disable it.
        if self.sync is None and getattr(args.command, 'connect', True):
            self.sync = ZODBSync(config=self.config, logger=logger)

        if self.config and not args.no_lock:
            self.lock = filelock.FileLock(
                os.path.join(self.config['base_dir'], '.zodbsync.lock')
            )

        self.command = args.command(
            args=self.args,
            logger=self.logger,
            config=self.config,
            sync=self.sync,
            lock=self.lock,
        )

        return self.command

    def run(self, *argv):
        """
        Parse arguments and run command
        """
        self.parse(*argv).run()



#!/usr/bin/env python

from ..subcommand import SubCommand


class Playback(SubCommand):
    '''Play back objects from the file system to the Data.FS'''
    @staticmethod
    def add_args(parser):
        parser.add_argument(
            '--override', '-o', action='store_true',
            help='Override object type changes when uploading',
            default=False
        )
        parser.add_argument(
            '--no-recurse', action='store_true',
            help='''Only upload metadata, do not remove elements or recurse.
            Note: If a path no longer present on the file system is given, it
            is still removed.''',
            default=False
        )
        parser.add_argument(
            '--skip-errors', action='store_true',
            help="Skip failed objects and continue",
            default=False
        )
        parser.add_argument(
            '--dry-run', action='store_true', default=False,
            help='Roll back at the end.',
        )
        parser.add_argument(
            'path', type=str, nargs='*',
            help='Sub-Path in Data.fs to be played back',
        )

    @SubCommand.with_lock
    def run(self):
        self.sync.playback_paths(
            paths=self.args.path,
            recurse=not self.args.no_recurse,
            override=self.args.override,
            skip_errors=self.args.skip_errors,
            dryrun=self.args.dry_run,
        )



import os
import tempfile
import shutil
import subprocess

'''
These classes together define an environment containing of a running ZEO
instance and a git repository (both running in temporary directories) as well
as a Zope instance configuration connecting to this ZEO and a ZODBSync
configuration connecting everything. They are used by the fixture defined in
conftest.py to provide an environment for the tests.
'''


class ZeoInstance():
    def __init__(self):
        self.path = tempfile.mkdtemp()
        subprocess.check_call(['mkzeoinstance', self.path])

        # replace address line to use a socket
        fname = self.path + '/etc/zeo.conf'
        with open(fname) as f:
            lines = f.readlines()
        subst = '  address ' + self.sockpath() + '\n'
        lines = [
            subst if '  address' in line else line
            for line in lines
        ]
        with open(fname, 'w') as f:
            f.writelines(lines)

        self.zeo = subprocess.Popen([self.path + '/bin/runzeo'])

    def sockpath(self):
        return self.path + '/var/zeo.sock'

    def cleanup(self):
        self.zeo.terminate()
        self.zeo.wait()
        shutil.rmtree(self.path)


class Repository():
    def __init__(self):
        self.path = tempfile.mkdtemp()
        commands = [
            ['init'],
            ['branch', '-m', 'autotest'],
            ['config', 'user.email', 'test@zodbsync.org'],
            ['config', 'user.name', 'testrepo'],
        ]
        for cmd in commands:
            subprocess.check_call(['git', '-C', self.path] + cmd)

    def cleanup(self):
        shutil.rmtree(self.path)


class ZopeConfig():
    def __init__(self, zeosock):
        self.path = tempfile.mkdtemp()
        self.config = self.path + '/zope.conf'
        content = '''
%define INSTANCE {path}
%define ZEO_SERVER {zeosock}

instancehome $INSTANCE

<zodb_db main>
    <zeoclient>
      server $ZEO_SERVER
      storage 1
      name zeostorage
      var $INSTANCE/var
      cache-size 20MB
    </zeoclient>
   mount-point /
</zodb_db>
        '''.format(zeosock=zeosock, path=self.path)

        with open(self.config, 'w') as f:
            f.write(content)

    def cleanup(self):
        shutil.rmtree(self.path)


class ZODBSyncConfig():
    def __init__(self, env):
        self.folder = tempfile.mkdtemp()
        os.mkdir(self.folder + '/layers')
        self.path = self.folder + '/zodb.py'
        with open(self.path, 'w') as f:
            f.write('''
conf_path = '{zopeconf}'
datafs_path = '{zeopath}/var/Data.fs'
manager_user = 'perfact'
create_manager_user = True
default_owner = 'perfact'
base_dir = '{repodir}'
commit_name = "Zope Developer"
commit_email = "zope-devel@example.de"
commit_message = "Generic commit message."
layers = "{root}/layers"
            '''.format(
                zopeconf=env['zopeconfig'].config,
                zeopath=env['zeo'].path,
                repodir=env['repo'].path,
                root=self.folder,
            ))

    def cleanup(self):
        shutil.rmtree(self.folder)


class JSLib():
    '''
    A test JS library containing some JS and CSS files
    '''
    def __init__(self):
        self.path = tempfile.mkdtemp()

        self.js_folder = os.path.join(self.path, 'js', 'plugins')
        os.makedirs(self.js_folder)
        with open(os.path.join(self.js_folder, 'something.js'), 'w') as jsfile:
            jsfile.write('alert(1);\n')

        self.css_folder = os.path.join(self.path, 'css', 'skins')
        os.makedirs(self.css_folder)
        with open(os.path.join(self.css_folder, 'dark.css'), 'w') as cssfile:
            cssfile.write('body { background-color: black; }\n')

        with open(os.path.join(self.path, 'ignoreme'), 'w') as ignorefile:
            ignorefile.write('something to ignore')

    def cleanup(self):
        shutil.rmtree(self.path)



#!/usr/bin/python3

import os

from ..subcommand import SubCommand
from ..helpers import StrRepr, literal_eval
from ..zodbsync import mod_format


class Reformat(SubCommand):
    """
    Rewrite commits from given commit to HEAD to post-4.3.2 formatting
    """
    # Note that in contrast to most subcommands, this should probably be
    # executed on a local repository not directly in sync with a ZODB.  No
    # rollback in case of error is implemented yet!

    connect = False

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            'commit', type=str,
            help="Starting point before first commit to rewrite"
        )

    def head(self):
        return self.gitcmd_output('rev-parse', 'HEAD').strip()

    @SubCommand.with_lock
    def run(self):
        start = self.args.commit
        commits_raw = self.gitcmd_output(
            'log', '--format=%H', '--reverse',
            '{}..HEAD'.format(start)
        )
        commits = [c for c in commits_raw.strip().split('\n') if c]

        self.gitcmd_run('reset', '--hard', start)
        base = self.config['base_dir']
        paths = []
        for root, dirs, files in os.walk(os.path.join(base, '__root__')):
            if '__meta__' in files:
                paths.append(os.path.join(root, '__meta__'))
        if self.reformat(paths):
            self.gitcmd_run('commit', '-a', '-m', 'zodbsync reformat')
        for idx, commit in enumerate(commits):
            print("Processing commit {}/{}".format(idx+1, len(commits)))
            cur = self.head()
            paths = list({
                os.path.join(base, line)
                for line in self.gitcmd_output(
                    'diff', '--name-only', '--no-renames', commit + '~', commit
                ).strip().split('\n')
                if line
            })
            metas = {path for path in paths if path.endswith('/__meta__')}
            if self.reformat(metas, True):
                self.gitcmd_run('commit', '-a', '-m', 'reverse')
            self.gitcmd_run('checkout', '--no-overlay', commit, '--', *paths)
            self.gitcmd_try('commit', '--no-edit', '-c', commit)
            self.reformat(metas)
            # Squash commits together with original message
            self.gitcmd_run('reset', cur)
            while paths:
                self.gitcmd_run('add', *paths[:100])
                del paths[:100]
            self.gitcmd_try('commit', '--no-edit', '-c', commit)

    def reformat(self, paths, legacy=False):
        changed = False
        for path in paths:
            if not os.path.exists(path):
                continue
            for _ in range(1000):  # Just in case we have no stdin
                with open(path) as f:
                    orig = f.read()
                try:
                    data = literal_eval(orig)
                except Exception:
                    # This allows the user to open the file manually, make it
                    # valid to be parsed by literal_eval and continue the
                    # process
                    print("Unable to parse path", path)
                    input("Retrying. Ctrl+C to cancel.")
                    continue
                break
            else:
                raise ValueError("Meta file could not be parsed")
            if legacy:
                fmt = StrRepr()(data, legacy=True)
            else:
                fmt = mod_format(data)
            if orig == fmt:
                continue
            with open(path, 'w') as f:
                f.write(fmt)
            changed = True

        return changed



#!/usr/bin/env python

import os
import sys
import warnings

from ..subcommand import SubCommand
from ..zodbsync import mod_format


PY2 = (sys.version_info.major == 2)


def create_template(type, content_type=None):
    result = {'type': type, 'title': ''}
    if content_type is not None:
        result['props'] = [[
                ('id', 'content_type'),
                ('type', 'string'),
                ('value', content_type)
        ]]
    return result


META_TEMPLATES = {
    'folder': create_template('Folder'),
    'js': create_template('File', 'application/javascript'),
    'css': create_template('File', 'text/css'),
    'html': create_template('File', 'text/html'),
    'odt': create_template('File', 'application/vnd.oasis.opendocument.text'),
    'ods': create_template(
        'File', 'application/vnd.oasis.opendocument.spreadsheet'),
    'pdf': create_template('File', 'application/pdf'),
    'svg': create_template('File', 'image/svg+xml'),
}
META_DEFAULT = create_template('File', 'application/octet-stream')


class Upload(SubCommand):
    '''[DEPRECATED] Upload a folder structure, e.g. a JS library, to zope
    Data.FS'''

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            'source', type=str,
            help='Path of source library folder',
        )
        parser.add_argument(
            'path', type=str,
            help='Sub-Path in Data.fs to put source folder',
        )
        parser.add_argument(
            '--override', '-o', action='store_true',
            help='Override object type changes when uploading',
            default=False
        )
        parser.add_argument(
            '--skip-errors', action='store_true',
            help="Skip failed objects and continue",
            default=False
        )
        parser.add_argument(
            '--dry-run', action='store_true', default=False,
            help='Roll back at the end.',
        )
        parser.add_argument(
            '--replace-periods', action='store_true', default=False,
            help='Replace periods in file names with underscores',
        )
        parser.add_argument(
            '--valid-extensions', type=str,
            help=(
                'Only upload files with the extensions listed '
                '(comma separated list). Allow all extensions by default.'
            )
        )

    @SubCommand.with_lock
    def run(self):
        '''
        Convert source folder into zodbsync compatible struct in repodir
        and upload it.
        '''
        warnings.warn(
            'Static external assets should not be included in the Data.FS',
            DeprecationWarning
        )
        self.check_repo()

        # we need both filesystem and Data.fs path representation
        data_fs_path, filesystem_path = self.datafs_filesystem_path(
            self.args.path
        )

        valid_extensions = self.args.valid_extensions
        if valid_extensions:
            # Parse comma separated list
            valid_extensions = [
                item.strip()
                for item in valid_extensions.split(',')
                if len(item.strip()) > 0
            ]

        # conversion loop: iterate over source folder, create folders in
        # repodir and corresponding files
        for cur_dir_path, dirs, files in os.walk(self.args.source):
            # relative path to be created in repodir
            cur_dir = os.path.relpath(cur_dir_path, self.args.source)

            # repodir folder creation
            new_folder = os.path.join(filesystem_path, cur_dir)
            if PY2:
                os.makedirs(new_folder)
            else:
                os.makedirs(new_folder, exist_ok=True)

            # do not forget meta file for folder
            self.create_file(
                file_path=os.path.join(new_folder, '__meta__'),
                content=mod_format(META_TEMPLATES['folder'])
            )

            # now check files inside of folder
            for filename in files:
                file_ending = filename.split('.')[-1]

                # bail out if not a valid extension
                if (valid_extensions is not None and
                        file_ending not in valid_extensions):
                    continue

                # read file content from source file
                with open(
                            os.path.join(cur_dir_path, filename), 'rb'
                        ) as sourcefile:
                    file_content = sourcefile.read()

                # choose the original filename, or replace periods
                if self.args.replace_periods:
                    repo_filename = filename.replace('.', '_')
                else:
                    repo_filename = filename

                # in repo each file gets its own folder ...
                new_file_folder = os.path.join(
                    new_folder, repo_filename
                )
                os.makedirs(new_file_folder)

                # ... containing __meta__ and __source__ file
                self.create_file(
                    file_path=os.path.join(new_file_folder, '__meta__'),
                    content=mod_format(
                        META_TEMPLATES.get(file_ending, META_DEFAULT)
                    )
                )
                self.create_file(
                    file_path=os.path.join(
                        new_file_folder, '__source__.' + file_ending
                    ),
                    content=file_content,
                    binary=True
                )

        # conversion done, start playback
        try:
            self.sync.playback_paths(
                paths=[data_fs_path],
                recurse=True,
                override=self.args.override,
                skip_errors=self.args.skip_errors,
                dryrun=self.args.dry_run,
            )

            if self.args.dry_run:
                self.abort()

        except Exception:
            self.logger.exception('Error uploading files. Resetting.')
            self.abort()
            raise



#!/usr/bin/env python

from ..subcommand import SubCommand


class Pick(SubCommand):
    '''Cherry-pick commits, apply them and play back affected objects'''
    @staticmethod
    def add_args(parser):
        parser.add_argument(
            '--skip-errors', action='store_true', default=False,
            help='Skip failed objects and continue',
        )
        parser.add_argument(
            '--dry-run', action='store_true', default=False,
            help='Only check for conflicts and roll back at the end.',
        )
        parser.add_argument(
            '--grep', type=str, help="""Find commits starting from the given
            ones, limiting to those with commit messages matching the
            pattern - like "git log --grep".""",
        )
        parser.add_argument(
            'commit', type=str, nargs='*',
            help='''Commits that are checked for compatibility and applied,
            playing back all affected paths at the end.'''
        )

    @SubCommand.gitexec
    def run(self):
        commits = []
        if self.args.grep:
            commits = self.gitcmd_output(
                'log', '--grep', self.args.grep,
                '--format=%H', '--reverse', *self.args.commit
            ).split('\n')
        else:
            for commit in self.args.commit:
                if '..' in commit:
                    # commit range
                    commits.extend(self.gitcmd_output(
                        'log', '--format=%H', '--reverse', commit
                    ).split('\n'))
                else:
                    commits.append(commit)

        for commit in commits:
            if not commit:
                continue
            self.logger.info('Checking and applying %s.' % commit)
            # capture output and discard so we don't clutter stdout
            # Python 2 has no subprocess.DEVNULL.
            self.gitcmd_output('cherry-pick', '--strategy', 'resolve', commit)



#!/usr/bin/env python

from ..subcommand import SubCommand


class Checkout(SubCommand):
    '''Switch to another branch'''
    @staticmethod
    def add_args(parser):
        parser.add_argument(
            '--skip-errors', action='store_true', default=False,
            help='Skip failed objects and continue',
        )
        parser.add_argument(
            '--dry-run', action='store_true', default=False,
            help='Only check for conflicts and roll back at the end.',
        )
        parser.add_argument(
            '-b', action='store_true', default=False,
            help='Create branch.',
        )
        parser.add_argument(
            '-t', '--track', type=str, help='Set up upstream configuration.'
        )
        parser.add_argument(
            '--reset', type=str,
            help='Reset branch onto given commit.',
        )
        parser.add_argument(
            '--rebase', type=str,
            help='Rebase branch onto given commit.',
        )
        parser.add_argument(
            'branch', type=str,
            help='''Branch name'''
        )

    @SubCommand.gitexec
    def run(self):
        self.logger.info('Checking out %s.' % self.args.branch)
        cmd = ['checkout']
        if self.args.b:
            cmd.append('-b')
        cmd.append(self.args.branch)
        if self.args.b and self.args.track:
            cmd.extend(['--track', self.args.track])
        self.gitcmd_run(*cmd)
        if self.args.reset:
            self.gitcmd_run('reset', '--hard', self.args.reset)
        if self.args.rebase:
            self.gitcmd_run('rebase', self.args.rebase)



#!/usr/bin/env python

import subprocess

from ..subcommand import SubCommand


class Exec(SubCommand):
    '''Execute a command and play back any paths changed between old and new
    HEAD'''
    @staticmethod
    def add_args(parser):
        parser.add_argument(
            '--skip-errors', action='store_true', default=False,
            help='Skip failed objects and continue',
        )
        parser.add_argument(
            '--dry-run', action='store_true', default=False,
            help='Only check for conflicts and roll back at the end.',
        )
        parser.add_argument(
            '--nocd', action='store_true', default=False,
            help='Do not cd to git repo for command',
        )
        parser.add_argument(
            'cmd', type=str, help='''command to be executed'''
        )

    @SubCommand.gitexec
    def run(self):
        subprocess.check_call(
            self.args.cmd,
            cwd=None if self.args.nocd else self.sync.base_dir,
            shell=True,
        )



#!/usr/bin/env python

import sys
import subprocess
import os

import filelock
import json

from .helpers import Namespace


class SubCommand(Namespace):
    '''
    Base class for different sub-commands to be used by zodbsync.
    '''

    # The presence of one of these in the .git folder indicates that some
    # process was not finished correctly, which is used to trigger a rollback
    # in some operations. Are these all?
    git_state_indicators = ['rebase-merge', 'rebase-apply', 'CHERRY_PICK_HEAD',
                            'MERGE_HEAD', 'REVERT_HEAD']

    @staticmethod
    def add_args(parser):
        ''' Overwrite to add arguments specific to sub-command. '''
        pass

    def acquire_lock(self, timeout=10):
        if self.args.no_lock:
            return
        try:
            self.lock.acquire(timeout=1)
        except filelock.Timeout:
            self.logger.debug("Acquiring exclusive lock...")
            try:
                self.lock.acquire(timeout=timeout-1)
            except filelock.Timeout:
                self.logger.error("Unable to acquire lock.")
                sys.exit(1)

    def release_lock(self):
        if not self.args.no_lock:
            self.lock.release()

    @staticmethod
    def with_lock(func):
        """
        Decorator for instance methods that are enveloped by a lock
        """
        def wrapper(self, *args, **kwargs):
            self.acquire_lock()
            try:
                result = func(self, *args, **kwargs)
            finally:
                self.release_lock()
            return result

        return wrapper

    def gitcmd(self, *args):
        # use "--no-pager" instead of "-P" for compatibility / readability
        return ['git', '--no-pager', '-C', self.config['base_dir']
                ] + list(args)

    def gitcmd_run(self, *args):
        '''Wrapper to run a git command.'''
        subprocess.check_call(self.gitcmd(*args))

    def gitcmd_try(self, *args):
        '''Wrapper to run a git command, ignoring return code.'''
        subprocess.call(self.gitcmd(*args))

    def gitcmd_output(self, *args):
        '''Wrapper to run a git command and return the output.'''
        return subprocess.check_output(
            self.gitcmd(*args), universal_newlines=True
        )

    def datafs_filesystem_path(self, path):
        '''Create absolute filesystem path from Data.fs path
        '''

        if path.startswith('./'):
            path = path[2:]

        if path.startswith('/'):
            path = path[1:]

        data_fs_path = path
        if path.startswith(self.sync.site):
            filesystem_path = os.path.join(
                self.config["base_dir"],
                path
            )
        else:
            filesystem_path = os.path.join(
                self.config["base_dir"],
                self.sync.site,
                path
            )

        return data_fs_path, filesystem_path

    def _branch_info(self):
        """Returns currently checked out branch as well as where each branch
        points."""
        current = self.gitcmd_output(
            'rev-parse', '--abbrev-ref', 'HEAD'
        ).strip()

        branches = {}  # branchname -> commitid
        output = self.gitcmd_output('show-ref', '--heads')
        for line in output.strip().split('\n'):
            commit, refname = line.split()
            refname = refname[len('refs/heads/'):]
            branches[refname] = commit

        return (current, branches)

    def check_repo(self):
        '''Check for unstaged changes and memorize current commit. Move
        unstaged changes away via git stash'''
        self.unstaged_changes = [
            line[3:]
            for line in self.gitcmd_output(
                'status', '--untracked-files', '-z'
            ).split('\0')
            if line
        ]

        self.orig_branch, self.branches = self._branch_info()

        if self.unstaged_changes:
            self.logger.warning(
                "Unstaged changes found. Moving them out of the way."
            )
            self.gitcmd_run('stash', 'push', '--include-untracked')

        # The commit to compare to with regards to changed files
        self.orig_commit = self.branches[self.orig_branch]

    def _playback_paths(self, paths):
        paths = self.sync.prepare_paths(paths)
        dryrun = self.args.dry_run

        playback_hook = self.config.get('playback_hook', None)
        if playback_hook and os.path.isfile(playback_hook):
            proc = subprocess.Popen(
                playback_hook, stdin=subprocess.PIPE,
                stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                universal_newlines=True)
            out, _ = proc.communicate(json.dumps({'paths': paths}))
            returncode = proc.returncode
            if returncode:
                raise AssertionError(
                    "Error calling playback hook, returncode "
                    "{}, [[{}]] on {}".format(
                        returncode, playback_hook, out
                    )
                )
            phases = json.loads(out)
        else:
            phases = [{'name': 'playback', 'paths': paths}]
            if self.config.get('run_after_playback', None):
                phases[-1]['cmd'] = self.config['run_after_playback']

        for ix, phase in enumerate(phases):
            phase_name = phase.get('name') or str(ix)
            phase_cmd = phase.get('cmd')

            self.sync.playback_paths(
                paths=phase['paths'],
                recurse=False,
                override=True,
                skip_errors=self.args.skip_errors,
                dryrun=dryrun,
            )

            if dryrun or not (phase_cmd and os.path.isfile(phase_cmd)):
                continue

            self.logger.info(
                'Calling phase %s, command: %s', phase_name, phase_cmd
            )
            proc = subprocess.Popen(
                phase_cmd, stdin=subprocess.PIPE,
                stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                universal_newlines=True)
            out, _ = proc.communicate(json.dumps(
                {'paths': phase['paths']}
            ))
            returncode = proc.returncode

            if returncode:
                self.logger.error(
                    "Error during phase command %s, %s",
                    returncode, out
                )
                if sys.stdin.isatty():
                    print("Enter 'y' to continue, other to rollback")
                    res = input()
                    if res == 'y':
                        continue

                raise AssertionError(
                    "Unrecoverable error in phase command"
                )
            else:
                self.logger.info(out)

    @staticmethod
    def gitexec(func):
        """
        Decorator for wrapping an operation with playback:
        - Stash unstaged changes away
        - memorize the current commit
        - do something
        - check for conflicts
        - play back changed objects (diff between old and new HEAD)
        - unstash
        """

        @SubCommand.with_lock
        def wrapper(self, *args, **kwargs):
            # Check for unstaged changes
            self.check_repo()

            try:
                self.paths = []
                func(self, *args, **kwargs)

                # Fail and roll back for any of the markers of an interrupted
                # git process (merge/rebase/cherry-pick/etc.)
                for fname in self.git_state_indicators:
                    path = os.path.join(self.sync.base_dir, '.git', fname)
                    assert not os.path.exists(path), "Git state not clean"

                files = {
                    line for line in self.gitcmd_output(
                        'diff', self.orig_commit, '--name-only', '--no-renames'
                    ).strip().split('\n')
                    if line
                }
                conflicts = files & set(self.unstaged_changes)
                assert not conflicts, "Change in unstaged files, aborting"

                # Make unique and sort
                self.paths = sorted(set(files))

                self._playback_paths(self.paths)

                if self.args.dry_run:
                    self.abort()
                elif self.unstaged_changes:
                    self.gitcmd_run('stash', 'pop')

            except Exception:
                self.logger.error('Error during operation. Resetting.')

                # Special handling in case of interrupted cherry-pick: show
                # differences in affected files
                cpfname = os.path.join(self.sync.base_dir,
                                       '.git/CHERRY_PICK_HEAD')
                if os.path.exists(cpfname):
                    with open(cpfname) as f:
                        failed_commit = f.read().strip()
                    output = self.gitcmd_output(
                        'diff-tree', '--no-commit-id', '--name-only',
                        '-r', failed_commit,
                    )
                    affected_files = [
                        line
                        for line in output.strip().split('\n')
                        if line
                    ]
                    self.logger.error("The cherry-pick failed due to the"
                                      " following difference:")
                    try:
                        self.gitcmd_run('diff', failed_commit + '~', 'HEAD',
                                        '--', *affected_files)
                    except subprocess.CalledProcessError:
                        # Make sure the call to abort is still done, even if
                        # for example the list of affected_files is too long
                        self.logger.exception("Unable to show diff")

                self.abort()
                # if we are not in dryrun we can't be sure we havent already
                # committed some stuff to the data-fs so playback all paths
                # abort
                if not self.args.dry_run and self.paths:
                    self.sync.playback_paths(
                        paths=self.paths,
                        recurse=False,
                        override=True,
                        skip_errors=True,
                        dryrun=False,
                    )
                raise

        return wrapper

    def create_file(self, file_path, content, binary=False):
        flags = 'wb' if binary else 'w'
        with open(file_path, flags) as create_file:
            create_file.write(content)

    def abort(self):
        '''Abort actions on repo and revert stash. check_repo must be
        called before this can be used'''
        current, branches = self._branch_info()
        # reset currently checked out branch
        target = self.branches.get(current)
        if target is None:
            # The branch was not originally present - we still need to reset it
            # to abort any operation
            target = branches[current]
        self.gitcmd_run('reset', '--hard', target)

        # reset all other branches
        for branch in self.branches:
            if branch == current:
                continue
            if branches[branch] == self.branches[branch]:
                continue
            self.gitcmd_run('branch', '-f', branch, self.branches[branch])

        # check out original branch
        if current != self.orig_branch:
            self.gitcmd_run('checkout', self.orig_branch)

        if self.unstaged_changes:
            self.gitcmd_run('stash', 'pop')

    def run(self):
        '''
        Overwrite for the action that is to be performed if this subcommand is
        chosen.
        '''
        print(self.args)



#!/usr/bin/env python

import os
import hashlib

from ..subcommand import SubCommand


class LayerHash(SubCommand):
    """Compute hashes for the contents of a layer"""
    subcommand = 'layer-hash'
    connect = False
    use_config = False

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            'path', type=str, help="Base dir of layer"
        )

    def run(self):
        """
        Create a sorted list of hashes for each folder below <path>/__root__.
        This is used when changing the contents of a layer to recognize which
        objects are to be played back.
        For each folder that contains files, it creates a sha1sum over:
        - The sorted list of files
        - The concatenation of the file contents
        The output is written to <path>/.checksums.
        """
        root = os.path.join(self.args.path, '__root__')
        todo = [root]
        with open(os.path.join(self.args.path, '.checksums'), 'w') as fd:
            while todo:
                path = todo.pop()
                entries = list(os.scandir(path))
                todo.extend(sorted((entry.path for entry in entries
                                    if entry.is_dir()), reverse=True))
                files = sorted(entry.path for entry in entries
                               if entry.is_file())
                if not files:
                    continue

                h = hashlib.sha1()
                for file in files:
                    h.update(file.encode('utf-8') + b'\n')
                h.update(b'\n')
                for fname in files:
                    with open(fname, 'rb') as f:
                        while data := f.read(1024*1024):
                            h.update(data)
                print(h.hexdigest(), path[len(root):] or '/', file=fd)



#!/usr/bin/env python

import subprocess

from ..subcommand import SubCommand


class WithLock(SubCommand):
    """Execute a shell command by first grabbing the lock"""
    subcommand = 'with-lock'
    connect = False

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            'cmd', type=str, help="Shell-command to be executed",
        )

    @SubCommand.with_lock
    def run(self):
        subprocess.check_call(self.args.cmd, shell=True)



from .zodbsync import mod_read, mod_write
from .extedit import launch as extedit_launch
from .helpers import obj_modtime, db_modtime

__all__ = [
    'mod_read',
    'mod_write',
    'obj_modtime',
    'db_modtime',
    'extedit_launch',
]



# -*- coding: utf-8 -*-
import pytest

from .. import helpers


def test_remove_redundant_paths():
    """
    Check that redundant subpaths are actually removed
    """
    paths = [
        '/test',
        '/test/sub',
        '/another',
    ]
    target = [
        '/another',
        '/test',
    ]
    helpers.remove_redundant_paths(paths)
    assert paths == target


def test_remove_redundant_paths_only_real_subpaths():
    """
    Check that paths are only recognized as redundant if they are actually
    subpaths, not if the last path component starts with the other.
    """
    paths = ['/test', '/test2']
    new_paths = paths[:]
    helpers.remove_redundant_paths(new_paths)
    assert paths == new_paths


def test_converters():
    """
    Several tests for to_* methods
    """
    for value in ['test', b'test']:
        assert helpers.to_bytes(value) == b'test'
        assert helpers.to_string(value) == 'test'
    assert helpers.to_string([1]) == '[1]'
    assert helpers.to_bytes([1]) == b'[1]'
    assert helpers.to_bytes(memoryview(b'test')) == b'test'


def test_StrRepr():
    """
    Check recursive version of str_repr with a typical configuration for what
    is split to occupy one line for each element, reproducing the shown
    formatting.
    """
    fmt = """
[
    ('content', [
        'a',
        'b',
    ]),
    ('owner', (['acl_users'], 'admin')),
    ('perms', [
        ('View', False, [
            'Role_1',
            'Role_2',
        ]),
    ]),
    ('props', [
        [('id', 'columns'), ('type', 'tokens'), ('value', (
            'a',
            'b',
            'c',
        ))],
        [('id', 'other'), ('type', 'lines'), ('value', (
            'x',
            'y',
            'z',
        ))],
        [('id', 'scalar'), ('type', 'string'), ('value', 'test')],
    ]),
]
    """.strip() + '\n'

    data = dict(helpers.literal_eval(fmt))
    rules = {
        'perms': [4],
        'props': [5],
    }
    assert fmt == helpers.StrRepr()(data, rules)


def test_StrReprLegacy():
    """
    Reproduce the shown formatting of StrRepr when using legacy mode
    """
    fmt = """
[
    ('content', [
        'a',
        'b',
        ]),
    ('owner', (['acl_users'], 'admin')),
    ('perms', [('View', False, ['Role_1', 'Role_2'])]),
    ('props', [
        [('id', 'columns'), ('type', 'tokens'), ('value', ('a', 'b', 'c'))],
        [('id', 'other'), ('type', 'lines'), ('value', ('x', 'y', 'z'))],
        [('id', 'scalar'), ('type', 'string'), ('value', 'test')],
        ]),
]
    """.strip() + '\n'
    data = dict(helpers.literal_eval(fmt))
    assert fmt == helpers.StrRepr()(data, legacy=True)


def test_literal_eval():
    tests = [
        ["b'test'", b'test'],
        ["{1: 2}", {1: 2}],
        ["[1, 2, 3]", [1, 2, 3]],
        ["None", None],
    ]
    for orig, compare in tests:
        assert helpers.literal_eval(orig) == compare
    assert helpers.literal_eval("1 + 2") == 3
    assert helpers.literal_eval("-True") == -1
    with pytest.raises(Exception):
        helpers.literal_eval('f(1)')



#!/usr/bin/env python
from . import helpers
from .object_mixins import MixinModObj


class ModObj(object):
    meta_types = []

    @staticmethod
    def create(obj, data, obj_id):
        return

    @staticmethod
    def read(obj):
        return ()

    @staticmethod
    def write(obj, data):
        return

    @classmethod
    def collect_handlers(cls):
        """
        Create a dictionary mapping the supported meta_types of each class to
        the handler class by recursing into subclasses.
        """
        result = {
            meta_type: cls
            for meta_type in cls.meta_types
        }
        for sub in cls.__subclasses__():
            result.update(sub.collect_handlers())
        return result


class UserFolderObj(ModObj):
    meta_types = ['User Folder', ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct['OFSP'].manage_addUserFolder()

    @staticmethod
    def read(obj):
        users = []
        for user in obj.getUsers():
            users.append((
                user.getUserName(),
                user._getPassword(),
                user.roles,
                user.getDomains(),
            ))
        return {'users': users}

    @staticmethod
    def write(obj, data):
        users = obj.getUsers()
        current_users = [user.getUserName() for user in users]
        target_users = [user[0] for user in data['users']]
        obj._doDelUsers([u for u in current_users if u not in target_users])
        for user in data['users']:
            # according to AccessControl/userfolder.py, an existing user of the
            # same name is simply overwritten by _doAddUser
            obj._doAddUser(
                user[0],  # username
                '',  # password is set separately
                user[2],  # roles
                user[3],  # domains
            )
            # _doAddUser encrypts the given password, but the password in the
            # dump is already encrypted, so we have to set it manually
            obj.getUserById(user[0]).__ = user[1]


class DTMLDocumentObj(ModObj):
    meta_types = ['DTML Document', ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct['OFSP'].manage_addDTMLDocument(id=obj_id)

    @staticmethod
    def read(obj):
        return {
            'source': helpers.to_bytes(obj.raw)
        }

    @staticmethod
    def write(obj, data):
        obj.manage_edit(
            data=helpers.to_string(data['source']),
            title=data['title'],
        )


class DTMLMethodObj(DTMLDocumentObj):
    meta_types = ['DTML Method', ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct['OFSP'].manage_addDTMLMethod(id=obj_id)


class DTMLTeXObj(DTMLDocumentObj):
    meta_types = ['DTML TeX', ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct['DTMLTeX'].manage_addDTMLTeX(id=obj_id)


class ZForceObj(ModObj):
    meta_types = ['ZForce', ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct['ZForce'].manage_addZForce(
            id=obj_id,
            title='',
            query_id='',
            fields_id='',
        )

    @staticmethod
    def read(obj):
        return {}

    @staticmethod
    def write(obj, data):
        obj.manage_changeProperties(title=data['title'])


class ZSQLMethodObj(ModObj):
    meta_types = ['Z SQL Method', ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct['ZSQLMethods'].manage_addZSQLMethod(
            id=obj_id,
            title=data['title'],
            connection_id=data['connection_id'],
            arguments=data['args'],
            template=helpers.to_string(data['source']),
        )

    @staticmethod
    def read(obj):
        return {
            'args': obj.arguments_src,
            'connection_id': obj.connection_id,
            'source': helpers.to_bytes(obj.src),
            'advanced': sorted([
                ('connection_hook', obj.connection_hook),
                ('max_rows', obj.max_rows_),
                ('max_cache', obj.max_cache_),
                ('cache_time', obj.cache_time_),
                ('class_name', obj.class_name_),
                ('class_file', obj.class_file_),
            ]),
        }

    @staticmethod
    def write(obj, data):
        obj.manage_edit(
            title=data['title'],
            connection_id=data['connection_id'],
            arguments=data['args'],
            template=helpers.to_string(data['source']),
        )

        # Advanced settings
        adv = dict(data['advanced'])
        obj.manage_advanced(**adv)


class ExternalMethodObj(ModObj):
    meta_types = ['External Method', ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct['ExternalMethod'].manage_addExternalMethod(
            id=obj_id,
            title=data['title'],
            module=data['module'],
            function=data['function'],
        )

    @staticmethod
    def read(obj):
        return {
            'function': obj.function(),
            'module': obj.module(),
        }

    @staticmethod
    def write(obj, data):
        obj.manage_edit(
            title=data['title'],
            module=data['module'],
            function=data['function'],
        )


class FileObj(ModObj):
    meta_types = ['File', ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct['OFSP'].manage_addFile(id=obj_id)

    @staticmethod
    def read(obj):
        # XXX Precondition
        # Read chunked source from File/Image objects.
        return {'source': helpers.read_pdata(obj)}

    @staticmethod
    def write(obj, data):
        pd = helpers.prop_dict(data)

        # XXX Precondition?
        obj.manage_edit(
            filedata=data['source'],
            content_type=pd['content_type'],
            title=data['title'],
        )


class ImageObj(FileObj):
    meta_types = ['Image', ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct['OFSP'].manage_addImage(id=obj_id, file='')


class FolderObj(ModObj):
    meta_types = ['Folder', ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct['OFSP'].manage_addFolder(id=obj_id)

    @staticmethod
    def read(obj):
        # Site Access
        try:
            get_ar = obj.manage_addProduct['SiteAccess'].manage_getAccessRule
        except (AttributeError, KeyError):
            get_ar = None
        if get_ar:
            accessrule = get_ar and get_ar()
            if accessrule:
                return {'accessrule': accessrule}

        return {}

    @staticmethod
    def write(obj, data):
        obj.manage_changeProperties(title=data['title'])

        # Access Rule
        accessrule = data.get('accessrule', None)
        if accessrule:
            obj.manage_addProduct['SiteAccess'].manage_addAccessRule(
                method_id=accessrule
            )


class FolderOrderedObj(FolderObj):
    meta_types = ['Folder (Ordered)', ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct['OFSP'].manage_addOrderedFolder(id=obj_id)

    @staticmethod
    def read(obj):
        result = FolderObj.read(obj)
        # ordered folders store their contents to represent the ordering
        result['contents'] = [a[0] for a in obj.objectItems()]
        return result

    @staticmethod
    def write(obj, data):
        obj.manage_changeProperties(title=data['title'])

        # Access Rule
        accessrule = data.get('accessrule', None)
        if accessrule:
            obj.manage_addProduct['SiteAccess'].manage_addAccessRule(
                method_id=accessrule
            )

    @staticmethod
    def fix_order(obj, data):
        # sort children for ordered folders
        contents = data.get('contents', [])
        srv_contents = [a[0] for a in obj.objectItems()]

        # only use contents that are present in the object
        contents = [a for a in contents if a in srv_contents]
        obj.moveObjectsByDelta(contents, -len(contents))


class PageTemplateObj(ModObj):
    meta_types = ['Page Template', ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct['PageTemplates'].manage_addPageTemplate(
            id=obj_id,
            text=''
        )

    @staticmethod
    def read(obj):
        return {'source': obj._text}

    @staticmethod
    def write(obj, data):
        obj.pt_setTitle(data['title'], 'utf-8')
        obj.write(data['source'])


class RAMCacheManagerObj(ModObj):
    meta_types = ['RAM Cache Manager', ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct[
            'StandardCacheManagers'
        ].manage_addRAMCacheManager(id=obj_id)

    @staticmethod
    def read(obj):
        return {
            'settings': sorted(obj.getSettings().items())
        }

    @staticmethod
    def write(obj, data):
        obj.manage_editProps(
            title=data['title'],
            settings=dict(data['settings']),
        )


class AcceleratedHTTPCacheManagerObj(RAMCacheManagerObj):
    meta_types = ['Accelerated HTTP Cache Manager', ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct[
            'StandardCacheManagers'
        ].manage_addAcceleratedHTTPCacheManager(id=obj_id)


class ScriptPythonObj(ModObj):
    meta_types = ['Script (Python)', ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct['PythonScripts'].manage_addPythonScript(
            id=obj_id
        )

    @staticmethod
    def read(obj):
        return {
            'bindings': sorted(
                obj.getBindingAssignments().getAssignedNames().items()
            ),
            'args': obj.params(),
            'source': helpers.to_bytes(obj.body()),
            'proxy_roles': sorted(list(obj._proxy_roles)),
        }

    @staticmethod
    def write(obj, data):
        obj.ZPythonScript_setTitle(title=data['title'])
        obj.ZPythonScript_edit(params=data['args'],
                               body=helpers.to_string(data['source']))
        obj.ZBindings_edit(mapping=dict(data['bindings']))
        obj.manage_proxy(roles=data['proxy_roles'])


class ZPsycopgDAObj(ModObj):
    meta_types = ['Z Psycopg 2 Database Connection',
                  'Z Psycopg Database Connection', ]

    @staticmethod
    def create(obj, data, obj_id):
        # id, title, connection_string, check, zdatetime, tilevel, autocommit,
        # encoding
        obj.manage_addProduct['ZPsycopgDA'].manage_addZPsycopgConnection(
            id=obj_id,
            title=data['title'],
            connection_string=data['connection_string'],
        )

    @staticmethod
    def read(obj):
        # late additions may not yet be everywhere in the Data.fs
        obj_dict = {
            'autocommit': getattr(obj, 'autocommit', False),
            'readonlymode': getattr(obj, 'readonlymode', False),
            'connection_string': obj.connection_string,
            'encoding': obj.encoding,
            'tilevel': obj.tilevel,
            'zdatetime': obj.zdatetime,
        }
        # Place additional parameters into the object dict only if
        # they're set to non-default values
        if hasattr(obj, 'use_tpc') and obj.use_tpc:
            obj_dict['use_tpc'] = obj.use_tpc
        if hasattr(obj, 'datetime_str') and obj.datetime_str:
            obj_dict['datetime_str'] = obj.datetime_str
        return obj_dict

    @staticmethod
    def write(obj, data):
        parameters = {
            'title': data['title'],
            'connection_string': data['connection_string'],
            'zdatetime': data['zdatetime'],
            'tilevel': data['tilevel'],
            'autocommit': data.get('autocommit', False),
            'readonlymode': data.get('readonlymode', False),
            'encoding': data['encoding'],
        }
        if hasattr(obj, 'use_tpc'):
            parameters['use_tpc'] = data.get('use_tpc', False)
        if hasattr(obj, 'datetime_str'):
            parameters['datetime_str'] = data.get('datetime_str', False)
        obj.manage_edit(**parameters)


class ZPyODBCDAObj(ModObj):
    meta_types = ['Z PyODBC Database Connection', ]

    @staticmethod
    def create(obj, data, obj_id):
        # id, title, connection_string, check, zdatetime, tilevel, autocommit,
        # encoding
        obj.manage_addProduct['ZPyODBCDA'].addpyodbcConnectionBrowser(
            id=obj_id,
            title=data['title'],
            connection_string=data['connection_string'],
            auto_commit=data['autocommit'],
            MaxRows=data['maxrows'],
        )

    @staticmethod
    def read(obj):
        return {
            'autocommit': obj.auto_commit,
            'connection_string': obj.connx_string,
            'maxrows': obj.MaxRows,
        }

    @staticmethod
    def write(obj, data):
        obj.manage_edit(
            title=data['title'],
            connection_string=data['connection_string'],
            auto_commit=data['autocommit'],
            MaxRows=data['maxrows'],
        )


class ZcxOracleDAObj(ModObj):
    meta_types = ['Z cxOracle Database Connection', ]

    @staticmethod
    def create(obj, data, obj_id):
        # id, title, connection_string, check, zdatetime, tilevel, autocommit,
        # encoding
        obj.manage_addProduct['ZcxOracleDA'].manage_addZcxOracleConnection(
            id=obj_id,
            title=data['title'],
            connection_string=data['connection_string'],
        )

    @staticmethod
    def read(obj):
        return {
            'connection_string': obj.connection_string,
        }

    @staticmethod
    def write(obj, data):
        obj.manage_edit(
            title=data['title'],
            connection_string=data['connection_string'],
        )


class ZsapdbDAObj(ModObj):
    meta_types = ['Z sap Database Connection', ]

    @staticmethod
    def create(obj, data, obj_id):
        # id, title, connection_string, check, zdatetime, tilevel, autocommit,
        # encoding
        obj.manage_addProduct['ZsapdbDA'].manage_addZsapdbConnection(
            id=obj_id,
            title=data['title'],
            connection_string=data['connection_string'],
        )

    @staticmethod
    def read(obj):
        return {
            'connection_string': obj.connection_string,
        }

    @staticmethod
    def write(obj, data):
        obj.manage_edit(
            title=data['title'],
            connection_string=data['connection_string'],
        )


class SimpleUserFolderObj(FolderObj):
    meta_types = ['Simple User Folder', ]

    @staticmethod
    def create(obj, data, obj_id):
        obj.manage_addProduct['SimpleUserFolder'].addSimpleUserFolder()


class MailHostObj(ModObj):
    meta_types = ['Mail Host', ]

    @staticmethod
    def create(obj, data, obj_id):
        # id, title, connection_string, check, zdatetime, tilevel, autocommit,
        # encoding
        obj.manage_addProduct['MailHost'].manage_addMailHost(
            id=obj_id,
            title=data['title'],
            smtp_host=data['smtp_host'],
            smtp_port=data['smtp_port'],
        )

    @staticmethod
    def read(obj):
        return {
            'smtp_host': obj.smtp_host,
            'smtp_port': obj.smtp_port,
            'smtp_uid': obj.smtp_uid,
            'smtp_pwd': obj.smtp_pwd,
            'force_tls': obj.force_tls,
            'smtp_queue': obj.smtp_queue,
            'smtp_queue_directory': obj.smtp_queue_directory,
        }

    @staticmethod
    def write(obj, data):
        obj.manage_makeChanges(
            title=data['title'],
            smtp_host=data['smtp_host'],
            smtp_port=data['smtp_port'],
            smtp_uid=data.get('smtp_uid', ''),
            smtp_pwd=data.get('smtp_pwd', ''),
            force_tls=data.get('force_tls', False),
            smtp_queue=data.get('smtp_queue', ''),
            smtp_queue_directory=data.get('smtp_queue_directory', ''),
        )


object_handlers = ModObj.collect_handlers()


def mod_implemented_handlers(obj, meta_type):
    """
    Return the handlers of interfaces that the object implements. This always
    includes the one defined from the meta_type and may also include some of
    the mixins.
    """
    return [object_handlers[meta_type]] + [
        cls for cls in MixinModObj.__subclasses__()
        if cls.implements(obj)
    ]



#!/usr/bin/env python

import os
import shutil

from ..subcommand import SubCommand


class LayerInit(SubCommand):
    """Register layers by copying over the checksum files."""
    subcommand = 'layer-init'

    @staticmethod
    def add_args(parser):
        pass

    def run(self):
        tgt = os.path.join(self.sync.base_dir, '.layer-checksums')
        if not os.path.isdir(tgt):
            os.mkdir(tgt)

        for layer in self.sync.layers:
            ident = layer['ident']
            if not ident:
                continue
            src = os.path.join(layer['base_dir'], '.checksums')
            if os.path.exists(src):
                shutil.copy(src, os.path.join(tgt, ident))



#!/usr/bin/env python

import os
import shutil

from ..subcommand import SubCommand
# from ..helpers import hashdir


class LayerUpdate(SubCommand):
    """Update a layer. Check stored checksum file against file in layer and
    playback relevant paths."""
    subcommand = 'layer-update'

    @staticmethod
    def add_args(parser):
        parser.add_argument(
            '--dry-run', action='store_true', default=False,
            help='Only check for conflicts and roll back at the end.',
        )
        parser.add_argument(
            '--skip-errors', action='store_true',
            help="Skip failed objects and continue",
            default=False
        )
        parser.add_argument(
            'ident', type=str, nargs='*',
            help='Layer identifier(s)',
        )

    @staticmethod
    def read_checksums(fname):
        result = []
        for line in open(fname):
            line = line.rstrip('\n')
            if line:
                checksum, path = line.split(' ', 1)
                result.append((path, checksum))
        return result

    def run(self):
        paths = set()
        layer_paths = {layer['ident']: layer['base_dir']
                       for layer in self.sync.layers}
        fnames = []
        for ident in self.args.ident:
            assert ident in layer_paths, "Invalid ident"
            fnames.append((
                os.path.join(self.sync.base_dir, '.layer-checksums', ident),
                os.path.join(layer_paths[ident], '.checksums')
            ))
            old = self.read_checksums(fnames[-1][0])
            new = self.read_checksums(fnames[-1][1])
            oldidx = 0
            newidx = 0
            # Iterate through results, which are ordered by path. Add any
            # deviation to paths
            while oldidx < len(old) or newidx < len(new):
                if old[oldidx] == new[newidx]:
                    oldidx += 1
                    newidx += 1
                    continue
                oldpath = old[oldidx][0]
                newpath = new[newidx][0]
                if oldpath <= newpath:
                    paths.add(oldpath)
                    oldidx += 1
                    continue
                if newpath <= oldpath:
                    paths.add(newpath)
                    newidx += 1

        if not paths:
            return

        self._playback_paths(sorted(paths))

        if not self.args.dry_run:
            self.sync.record(sorted(paths), recurse=False, skip_errors=True,
                             ignore_removed=True)
            for path in paths:
                if self.sync.fs_pathinfo(path)['layeridx'] == 0:
                    self.logger.warning(
                        'Conflict with object in custom layer: ' + path
                    )
            for oldfname, newfname in fnames:
                shutil.copyfile(newfname, oldfname)
